/*
 * Copyright (C) 2001-2017 Jacek Sieka, arnetheduck on gmail point com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdinc.h"
#include "MappingManager.h"

#include "ConnectionManager.h"
#include "ConnectivityManager.h"
#include "format.h"
#include "LogManager.h"
#include "Mapper_MiniUPnPc.h"
#include "Mapper_NATPMP.h"
#include "ScopedFunctor.h"
#include "SearchManager.h"
#include "version.h"
#include <cstdint>

namespace dcpp {

MappingManager::MappingManager() : busy(false), renewal(0) {
	addMapper<Mapper_NATPMP>();
	addMapper<Mapper_MiniUPnPc>();
}

StringList MappingManager::getMappers() const {
	StringList ret;
	for(auto i = mappers.cbegin(), iend = mappers.cend(); i != iend; ++i)
		ret.push_back(i->first);
	return ret;
}

bool MappingManager::open() {
	if(getOpened())
		return false;

	if(mappers.empty()) {
		log(_("No port mapping interface available"));
		return false;
	}

	if(busy.test_and_set()) {
		log(_("Another port mapping attempt is in progress..."));
		return false;
	}

	start();

	return true;
}

void MappingManager::close() {
	join();

	if(renewal) {
		renewal = 0;
		TimerManager::getInstance()->removeListener(this);
	}

	if(working.get()) {
		close(*working);
		working.reset();
	}
}

bool MappingManager::getOpened() const {
	return working.get();
}

string MappingManager::getStatus() const {
	if(working.get()) {
		auto& mapper = *working;
		return string(F_("Successfully created port mappings on the "+deviceString(mapper)+" device with the "+mapper.getName()+" interface")  );
	}
	return _("Failed to create port mappings");
}

int MappingManager::run() {

	// cache ports
	uint16_t	conn_port = ConnectionManager::getInstance()->getPort();
	uint16_t	secure_port = ConnectionManager::getInstance()->getSecurePort();
	uint16_t	search_port = SearchManager::getInstance()->getPort();

	if(renewal) {
		Mapper& mapper = *working;

		ScopedFunctor([&mapper] { mapper.uninit(); });
		if(!mapper.init()) {
			// can't renew; try again later.
			renewLater(mapper);
			return 0;
		}

		auto addRule = [this, &mapper](const uint16_t& port, Mapper::Protocol protocol, const string& description) {
			// just launch renewal requests - don't bother with possible failures.
				mapper.open(Util::toString(port), protocol, string(F_(string(APPNAME)+" "+description+" port ("+Util::toString(port)+" "+Mapper::protocols[protocol]+")")));
		};

		addRule(conn_port, Mapper::PROTOCOL_TCP, _("Transfer"));
		addRule(secure_port, Mapper::PROTOCOL_TCP, _("Encrypted transfer"));
		addRule(search_port, Mapper::PROTOCOL_UDP, _("Search"));

		renewLater(mapper);
		return 0;
	}

	// move the preferred mapper to the top of the stack.
	const string& setting = SETTING(MAPPER);
	for(auto i = mappers.begin(); i != mappers.end(); ++i) {
		if(i->first == setting) {
			if(i != mappers.begin()) {
				auto mapper = *i;
				mappers.erase(i);
				mappers.insert(mappers.begin(), mapper);
			}
			break;
		}
	}

	for(auto i = mappers.begin(); i != mappers.end(); ++i) {
		unique_ptr<Mapper> pMapper(i->second(Util::getLocalIp()));
		Mapper& mapper = *pMapper;

		ScopedFunctor([&mapper] { mapper.uninit(); });
		if(!mapper.init()) {
			log(string(F_("Failed to initalize the "+mapper.getName()+" interface")));
			continue;
		}

		auto addRule = [this, &mapper](const uint16_t& port, Mapper::Protocol protocol, const string& description) -> bool {
			if( !mapper.open(Util::toString(port), protocol, string(F_(string(APPNAME)+" "+description+" port ("+Util::toString(port)+" "+Mapper::protocols[protocol]+")")
				 )))
			{
				this->log(string(F_("Failed to map the "+description+" port ("+Util::toString(port)+" "+Mapper::protocols[protocol]+") with the "+mapper.getName()+" interface")));
				mapper.close();
				mapper.uninit();
				return false;
			}
			return true;
		};

		if(!(addRule(conn_port, Mapper::PROTOCOL_TCP, _("Transfer")) &&
			addRule(secure_port, Mapper::PROTOCOL_TCP, _("Encrypted transfer")) &&
			addRule(search_port, Mapper::PROTOCOL_UDP, _("Search"))))
			continue;

		log(string(F_("Successfully created port mappings (Transfers: "+Util::toString(conn_port)+", Encrypted transfers: "+Util::toString(secure_port) +", Search: "+Util::toString(search_port)+") on the "+deviceString(mapper)+" device with the "+mapper.getName()+" interface") ));

		working = move(pMapper);

		if(!CONNSETTING(NO_IP_OVERRIDE)) {
			string externalIP = mapper.getExternalIP();
			if(!externalIP.empty()) {
				ConnectivityManager::getInstance()->set(SettingsManager::EXTERNAL_IP, externalIP);
			} else {
				// no cleanup because the mappings work and hubs will likely provide the correct IP.
				log(_("Failed to get external IP"));
			}
		}

		ConnectivityManager::getInstance()->mappingFinished(mapper.getName());

		renewLater(mapper);
		break;
	}

	if(!getOpened()) {
		log(_("Failed to create port mappings"));
		ConnectivityManager::getInstance()->mappingFinished(Util::emptyString);
	}
	busy.clear();
	return 0;
}

void MappingManager::close(Mapper& mapper) {
	if(mapper.hasRules()) {
		bool ret = mapper.init() && mapper.close();
		mapper.uninit();
		log(ret ?
			string(F_("Successfully removed port mappings from the "+deviceString(mapper)+" device with the "+ mapper.getName()+" interface")) :
			string(F_("Failed to remove port mappings from the "+deviceString(mapper)+" device with the "+mapper.getName()+" interface") ));
	}
}

void MappingManager::log(const string& message) {
	ConnectivityManager::getInstance()->log(string(F_("Port mapping: ")+  message));
}

string MappingManager::deviceString(Mapper& mapper) const {
	string name(mapper.getDeviceName());
	if(name.empty())
		name = _("Generic");
	return '"' + name + '"';
}

void MappingManager::renewLater(Mapper& mapper) {
	uint32_t minutes = mapper.renewal();
	if(minutes) {
		bool addTimer = !renewal;
		renewal = GET_TICK() + std::max(minutes, 10u) * 60 * 1000;
		if(addTimer) {
			TimerManager::getInstance()->addListener(this);
		}

	} else if(renewal) {
		renewal = 0;
		TimerManager::getInstance()->removeListener(this);
	}
}

void MappingManager::on(TimerManagerListener::Minute, uint64_t tick) noexcept {
	if(tick >= renewal && !busy.test_and_set())
		start();
}

} // namespace dcpp
